"""This module contains utilities for working with Abstract Syntax Trees (ASTs)."""

from __future__ import annotations

from ast import (
    AST,
    AnnAssign,
    AsyncFunctionDef,
    BinOp,
    BitOr,
    ClassDef,
    Constant,
    Expr,
    FunctionDef,
    List,
    Load,
    Module,
    Name,
    Slice,
    Subscript,
    Tuple,
    arg,
    arguments,
    expr,
    fix_missing_locations,
    stmt,
    unparse,
    walk,
)
from typing import Any, Sequence, TypeAlias

from zen_generator.core.formatting import format_python_code
from zen_generator.core.type_system import (
    convert_asyncapi_to_python,
    convert_python_to_asyncapi,
)

AnnotationNode: TypeAlias = AST | Subscript | List | Name | BinOp | Constant | Tuple | Slice | None
SCHEMA_PREFIX = "#/components/schemas/"


def parse_type_annotation(
    ast_annotation: AnnotationNode,
) -> list[dict[str, Any]]:
    """Parse an AST type annotation node into a list of type descriptions.

    Args:
        ast_annotation: The AST annotation node to parse (Subscript, Name, BinOp, etc.)

    Returns:
        List of dictionaries describing the type structure, each containing:
        - slice: The container type (list, dict, etc.) or None
        - value: The contained type name or None
    """
    match ast_annotation:
        case Subscript():
            kind = ast_annotation.value.id if hasattr(ast_annotation.value, "id") else None
            if kind == "dict" and hasattr(ast_annotation.slice, "elts"):
                value = [
                    ast_annotation.slice.elts[0].id,
                    ast_annotation.slice.elts[1].id,
                ]
            elif kind == "list":
                if isinstance(ast_annotation.slice, (Tuple, Subscript)):
                    value = parse_type_annotation(ast_annotation.slice)
                elif hasattr(ast_annotation.slice, "id"):
                    value = ast_annotation.slice.id
                else:
                    value = None
            else:
                value = ast_annotation.slice.id if hasattr(ast_annotation.slice, "id") else None
            return [{"slice": kind, "value": value}]
        case List():
            return [{"slice": "list", "value": None}]
        case Name():
            return [{"slice": None, "value": ast_annotation.id}]
        case BinOp():
            left_type = parse_type_annotation(ast_annotation.left)
            right_type = parse_type_annotation(ast_annotation.right)
            return left_type + right_type
        case Constant() if ast_annotation.value is None:
            return [{"slice": None, "value": None}]
        case _:
            return []


def convert_annotations_to_asyncapi_schemas(
    input_annotation: list[dict[str, Any]],
) -> dict[str, Any]:
    """Convert a list of dictionaries (result of `convert_ast_annotation_to_dict`) to an AsyncAPI schema.

    Args:
        input_annotation: A list of dictionaries containing the annotation information.

    Returns:
        A dictionary containing the AsyncAPI schema.
    """
    required = True
    schema_items: list[dict[str, Any]] = []

    valid_annotations = [annotation for annotation in input_annotation if annotation["value"] is not None]

    if len(valid_annotations) < len(input_annotation):
        required = False

    for annotation in valid_annotations:
        if isinstance(annotation["value"], str):
            primitive_type = convert_python_to_asyncapi(annotation["value"])
            if primitive_type:
                value = primitive_type
                item_type = "type"
            else:
                value = f"{SCHEMA_PREFIX}{annotation['value']}"
                item_type = "$ref"
        else:
            value = "object"
            item_type = "type"

        if annotation["slice"] == "list":
            schema_items.append({"type": "array", "items": {item_type: value}})
        else:
            schema_items.append({item_type: str(value)})

    properties: dict[str, Any] = (
        {} if not schema_items else schema_items[0] if len(schema_items) == 1 else {"oneOf": schema_items}
    )
    return {"required": required, "properties": properties}


def generate_class_schema(node: ClassDef) -> dict[str, Any]:
    """Generate a dictionary containing the AsyncAPI schema for a ClassDef node.

    Args:
        node: The ClassDef node to generate the schema for.

    Returns:
        A dictionary containing the AsyncAPI schema.
    """
    try:
        base = node.bases[0]
        base_class = base.id if hasattr(base, "id") else "object"
    except (IndexError, AttributeError):
        base_class = "object"

    properties = {}
    required = []
    schema = {"type": "object", "base_class": base_class}
    for class_element in node.body:
        if isinstance(class_element, AnnAssign) and isinstance(class_element.target, Name):
            property_name = class_element.target.id
            items = parse_type_annotation(class_element.annotation)
            conv = convert_annotations_to_asyncapi_schemas(items)
            if conv.get("required"):
                required.append(property_name)
                del conv["required"]
            properties[property_name] = conv.get("properties")

    schema["required"] = required
    schema["properties"] = properties
    return schema


def generate_code_from_ast(ast_nodes: list[Any]) -> str:
    """Converts an AST tree to a formatted string of Python code.

    Args:
        ast_nodes: A list of AST nodes to convert.

    Returns:
        A formatted string of Python code.
    """
    module = Module(body=ast_nodes, type_ignores=[])
    raw_code = unparse(fix_missing_locations(module))
    return format_python_code(raw_code)


def generate_component_schemas(tree: AST | None) -> dict[str, Any]:
    """Generates a dictionary containing the component schemas from an AST tree.

    Args:
        tree: The AST tree to parse. If None, an empty dictionary is returned.

    Returns:
        A dictionary containing the component schemas.
    """
    result: dict[str, Any] = {}
    if not tree:
        return result

    for node in walk(tree):
        if isinstance(node, ClassDef):
            result[node.name] = generate_class_schema(node)
    return result


def get_component_schemas(source: dict[str, Any] | None) -> dict[str, Any] | None:
    """Gets the component schemas from an AsyncAPI dictionary.

    Args:
        source: The AsyncAPI dictionary. If None, None is returned.

    Returns:
        The component schemas as a dictionary or None if the source does not contain
        the component schemas.
    """
    source = source or {}
    try:
        return source["components"]["schemas"]
    except KeyError:
        return None


def generate_bin_op(
    values: Sequence[str | Name | Subscript | Constant | BinOp | None],
) -> Name | Subscript | Constant | BinOp | None:
    """Generates a binary operation from a sequence of values.

    Args:
        values: A sequence of values to generate the binary operation from.

    Returns:
        A Name, Subscript, Constant, BinOp or None representing the generated
        binary operation.
    """
    if not values:
        return None

    if len(values) == 1:
        value = values[0]
        if value is None:
            return Constant(value=None)
        elif isinstance(value, str):
            return Name(id=value, ctx=Load())
        else:
            return value
    else:
        left = generate_bin_op(values[:-1])
        right = generate_bin_op([values[-1]])
        if left is not None and right is not None:
            return BinOp(left=left, op=BitOr(), right=right)
        else:
            return left or right


def convert_asyncapi_property_to_ast_node(
    pro: dict[str, Any] | None,
) -> Name | Subscript | Constant | BinOp | None:
    """Converts an AsyncAPI property to an AST node.

    Args:
        pro: The AsyncAPI property to convert

    Returns:
        The AST node representing the property
    """
    match pro:
        case None:
            return Constant(value=None)
        case {"type": "array"}:
            return _generate_subscript_from_property(pro)
        case {"type": type_value}:
            return Name(id=convert_asyncapi_to_python(type_value), ctx=Load())
        case {"$ref": ref_value}:
            ref_name = ref_value.replace(SCHEMA_PREFIX, "")
            return Name(id=convert_asyncapi_to_python(ref_name), ctx=Load())
        case {"oneOf": one_of_values}:
            type_nodes = [convert_asyncapi_property_to_ast_node(one_of) for one_of in one_of_values]
            return generate_bin_op(type_nodes)
        case _:
            return Constant(value=None)


def _generate_subscript_from_property(pro: dict[str, Any]) -> Subscript:
    """enerate a Subscript node representing a list type from an AsyncAPI array property.

    Args:
        pro: Dictionary describing the array property, which may include
            a reference ($ref) or a type field in its "items" section.

    Returns:
        The AST Subscript representing a list of the derived item type

    """
    items = pro.get("items", {})

    if "$ref" in items:
        type_name = items.get("$ref").replace(SCHEMA_PREFIX, "")
    else:
        type_name = items.get("type", "")

    item_type = convert_asyncapi_to_python(type_name)

    return Subscript(
        value=Name(id="list", ctx=Load()),
        slice=Name(id=item_type, ctx=Load()),
        ctx=Load(),
    )


def create_ast_function_definition(
    function_name: str,
    function_arguments: Sequence[arg],
    docstring: str | None,
    return_annotation: Any,
    is_async: bool = False,
    decorator_list: list[expr] | None = None,
) -> FunctionDef | AsyncFunctionDef:
    """Create a new AST function definition node.

    Args:
        function_name: The name of the function
        function_arguments: The arguments of the function
        docstring: The docstring of the function
        return_annotation: The return annotation of the function
        is_async: Whether the function is async
        decorator_list: The list of decorators for the function

    Returns:
        The AST function definition node

    """
    decorator_list = decorator_list or []
    body: list[stmt] = [Expr(value=Constant(value=Ellipsis))]

    if docstring:
        body.insert(0, Expr(value=Constant(value=docstring)))

    func_def_class = AsyncFunctionDef if is_async else FunctionDef

    return func_def_class(
        name=function_name,
        args=arguments(
            posonlyargs=[],
            args=list(function_arguments),
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        ),
        body=body,
        decorator_list=decorator_list,
        returns=return_annotation,
    )



"""This module contains utilities for generating AsyncAPI documents."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from zen_generator.core.ast_utils import convert_annotations_to_asyncapi_schemas, generate_component_schemas
from zen_generator.core.io import parse_python_file_to_ast, save_yaml_file
from zen_generator.core.parsing import function_content_reader


def create_async_api_content(
    app_name: str,
    models_schema: dict[str, Any],
    api_description: str | None,
    functions_parsed: dict[str, Any],
):
    """Generate the AsyncAPI document from the provided models and functions content.

    The function takes in the application name, the models schema, the functions docstring
    and the functions parsed content and generates the AsyncAPI document.

    :param app_name: The name of the application
    :param models_schema: The models schema
    :param api_description: The api docstring
    :param functions_parsed: The functions parsed content
    :return: The generated AsyncAPI document
    """
    channels = {}
    operations = {}
    components = {
        "channels": {},
        "operations": {},
        "messages": {},
        "schemas": models_schema,
    }
    for func, content in functions_parsed.items():
        # channels
        channels[func] = {"$ref": f"#/components/channels/{func}"}

        # operations
        operations[func] = {"$ref": f"#/components/operations/{func}"}

        # components.channels
        components["channels"][func] = {
            "messages": {
                "request": {"$ref": f"#/components/messages/{func}_request"},
                "response": {"$ref": f"#/components/messages/{func}_response"},
            }
        }

        # components.operations
        components["operations"][func] = {
            "action": "receive",
            "description": content.get("request", {}).get("description", ""),
            "channel": {"$ref": f"#/channels/{func}"},
            "messages": [{"$ref": f"#/channels/{func}/messages/request"}],
            "reply": {
                "channel": {"$ref": f"#/channels/{func}"},
                "messages": [{"$ref": f"#/channels/{func}/messages/response"}],
            },
        }

        # components.messages
        components["messages"][f"{func}_request"] = content.get("request")
        components["messages"][f"{func}_response"] = content.get("response")

    async_api_content: dict[str, Any] = {
        "asyncapi": "3.0.0",
        "info": {
            "title": app_name,
            "version": "0.0.1",
            "description": api_description or "",
        },
        "channels": channels,
        "operations": operations,
        "components": components,
    }
    return async_api_content


def generate_asyncapi_from_files(models_file: Path, functions_file: Path, output_path: Path, app_name: str) -> None:
    """Generate an AsyncAPI document from the provided model and function definitions.

    Args:
        models_file (Path): The path to the file containing the model definitions.
        functions_file (Path): The path to the file containing the function definitions.
        output_path (Path): The path where the generated AsyncAPI document will be saved.
        app_name (str): The name of the application.

    Returns:
        None
    """
    models_ast = parse_python_file_to_ast(models_file)
    models_schema = generate_component_schemas(models_ast)

    functions_ast = parse_python_file_to_ast(functions_file)
    api_description, functions_parsed = function_content_reader(functions_ast)

    async_api_content = create_async_api_content(app_name, models_schema, api_description, functions_parsed)

    save_yaml_file(async_api_content, output_path, app_name)



"""This module contains utilities for generating Python code from AsyncAPI specifications."""

from __future__ import annotations

from ast import (
    AnnAssign,
    Assign,
    Attribute,
    BinOp,
    BitOr,
    Call,
    ClassDef,
    Constant,
    Expr,
    Import,
    ImportFrom,
    Load,
    Name,
    Pass,
    Store,
    alias,
    arg,
    expr,
    parse,
    stmt,
    unparse,
)
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Sequence, cast

from zen_generator.core.ast_utils import (
    convert_asyncapi_property_to_ast_node,
    create_ast_function_definition,
    get_component_schemas,
)
from zen_generator.core.io import load_yaml, save_python_file


@dataclass
class BasePythonGenerator:
    """Base class for Python generators.

    This class provides a basic interface for generating Python code from
    AsyncAPI specifications.

    Attributes:
        output_file (Path): The path to the generated Python file.
        models_file (Path): The path to the generated models file.
        override_base_class (str | None): The base class to override in the generated code.
        decorator_list (Sequence[expr]): A list of decorators to apply to the generated class.
        additional_imports (Sequence[stmt | ImportFrom]): Additional imports to include in the generated code.
        extra_assignments (Sequence[stmt]): Additional assignments to include in the generated code.
    """

    models_ast: list[stmt] = field(init=False, repr=False, default_factory=list)
    functions_ast: list[stmt] = field(init=False, repr=False, default_factory=list)
    extra_imports: Sequence[stmt | ImportFrom] = field(default_factory=list)
    extra_assignments: Sequence[stmt] = field(default_factory=list)
    override_base_class: str | None = None
    decorator_list: Sequence[expr] = field(default_factory=list)
    source_content: dict[str, Any] = field(init=False, repr=False, default_factory=dict)
    component_schemas: dict[str, Any] = field(init=False, repr=False, default_factory=dict)
    is_async: bool = False

    def __post__init__(self) -> None:
        """Generate Python files from an AsyncAPI specification.



        Returns:
            None
        """
        self.models_ast.extend(
            [
                ImportFrom(module="__future__", names=[alias(name="annotations")], level=0),
            ]
        )

    def generate_files_from_asyncapi(
        self,
        source_file: Path,
        models_file: Path,
        functions_file: Path,
        app_name: str,
        is_async: bool = False,
    ) -> None:
        """Generate Python files from an AsyncAPI specification.

        This method takes in the path to the AsyncAPI file, the path to the generated
        models file, the path to the generated functions file and the name of the
        application. Optionally, it takes in a boolean indicating whether the generated
        code should be asynchronous.

        Args:
            source_file: The path to the AsyncAPI file.
            models_file: The path to the generated models file.
            functions_file: The path to the generated functions file.
            app_name: The name of the application.
            is_async: Whether the generated code should be asynchronous.

        Returns:
            None
        """
        self.is_async = is_async
        self.load_asyncapi_content(source_file)
        self.load_component_schemas()

        self.generate_models_ast()
        save_python_file(self.models_ast, models_file)
        self.generate_function_ast(app_name, models_file.stem)
        save_python_file(self.functions_ast, functions_file)

    def load_component_schemas(self) -> None:
        self.component_schemas = get_component_schemas(self.source_content) or {}

    def load_asyncapi_content(self, source_file) -> None:
        self.source_content = load_yaml(source_file) or {}

    def _add_logger_setup(self, app_name: str) -> None:
        """Add logger setup to the module.

        This method adds a logger setup to the module, which creates a logger
        with the name of the application and sets up a basic configuration.

        Args:
            app_name (str): The name of the application.

        Returns:
            None
        """
        self.functions_ast.append(
            Assign(
                targets=[Name(id="logger", ctx=Store())],
                value=Call(
                    func=Attribute(
                        value=Name(id="logging", ctx=Load()),
                        attr="getLogger",
                        ctx=Load(),
                    ),
                    args=[Constant(value=app_name)],
                    keywords=[],
                ),
            )
        )

    def _add_docstring(self) -> None:
        """Add docstring to the module.

        This method adds a docstring to the module, which contains the description
        of the application.

        Returns:
            None
        """
        docstring = self.source_content.get("info", {}).get("description", "Test API description")
        if isinstance(docstring, str):
            self.functions_ast.append(Expr(value=Constant(value=docstring)))

    def _add_models_import(self, module_name: str = "models") -> None:
        """Add import statement for the models module.

        This method adds an import statement to the module, which imports all
        the models defined in the `models` module.

        Args:
            module_name (str): The name of the module to import. Defaults to
                "models".

        Returns:
            None
        """
        if self.component_schemas.items():
            names = [alias(name=f"{model}") for model in self.component_schemas]
            import_from = ImportFrom(
                module=f".{module_name}",
                names=names,
                level=0,
            )
            self.functions_ast.append(import_from)

    def generate_models_ast(self) -> None:
        """Generate the models as a sequence of AST nodes.

        This method generates the models as a sequence of AST nodes, which
        correspond to the classes defined in the `models` module. The classes are
        generated from the `components/schemas` field of the AsyncAPI document.

        Returns:
            None
        """
        if not self.component_schemas:
            return

        self.models_ast.extend(self.extra_imports)

        for class_name, schema in self.component_schemas.items():
            class_body: list[stmt] = []
            base_class_id = self.override_base_class or schema.get("base_class", "object")
            if schema.get("properties"):
                for prop_name, prop_value in schema["properties"].items():
                    annotation = convert_asyncapi_property_to_ast_node(prop_value)
                    if annotation is not None and prop_name not in schema.get("required", []):
                        annotation = BinOp(
                            left=cast(expr, annotation),
                            op=BitOr(),
                            right=Constant(value=None),
                        )
                    if annotation is not None:
                        class_body.append(
                            AnnAssign(
                                target=Name(id=prop_name, ctx=Store()),
                                annotation=cast(expr, annotation),
                                simple=1,
                            )
                        )
            else:
                class_body = [Pass()]

            klass = ClassDef(
                name=class_name,
                bases=[Name(id=base_class_id, ctx=Load())],
                body=class_body,
                decorator_list=[],
                keywords=[],
            )
            self.models_ast.append(klass)

    def generate_function_ast(
        self,
        app_name: str,
        module_name: str = "models",
        logger: bool = True,
    ) -> None:
        """Generate the functions as a sequence of AST nodes.

        This method generates the functions as a sequence of AST nodes, which
        correspond to the functions defined in the `functions` module. The functions
        are generated from the `components/messages` field of the AsyncAPI document.

        Args:
            app_name (str): The name of the application.
            module_name (str, optional): The name of the module that contains the
                models. Defaults to "models".
            logger (bool, optional): Whether to add logger setup code. Defaults to
                True.

        Returns:
            None
        """
        self._add_docstring()

        self.functions_ast.append(ImportFrom(module="__future__", names=[alias(name="annotations")], level=0))
        self.functions_ast.extend(self.extra_imports)

        if logger:
            self.functions_ast.append(Import(names=[alias(name="logging")]))

        self._add_models_import(module_name)
        self.functions_ast.extend(self.extra_assignments)

        if logger:
            self._add_logger_setup(app_name)
        self._add_function_definitions()

    def _add_function_definitions(self) -> None:
        """Generate the functions as a sequence of AST nodes.

        This method generates the functions as a sequence of AST nodes, which
        correspond to the functions defined in the `functions` module. The functions
        are generated from the `components/messages` field of the AsyncAPI document.

        Args:
            components (dict): The components field of the AsyncAPI document.
            functions (dict): The functions field of the AsyncAPI document.

        Returns:
            None
        """
        components = self.source_content.get("components", {})
        functions = components.get("operations", {})

        if not functions:
            return

        for func_name in functions:
            processed_decorators = self._process_decorators(func_name)
            function_args = self._build_function_args(components, func_name)
            returns_node = self._build_return_annotation(components, func_name)
            description = functions[func_name].get("description")

            func_def = create_ast_function_definition(
                func_name, function_args, description, returns_node, self.is_async, processed_decorators
            )
            self.functions_ast.append(func_def)

    def _process_decorators(self, func_name: str) -> list[expr]:
        """Process decorators for a function.

        Process the decorator list for a function by replacing the `{func_name}`
        placeholder with the actual function name and converting the resulting
        string back to an AST node.

        Args:
            func_name (str): The name of the function.

        Returns:
            list[expr]: The processed decorator nodes.
        """
        processed_decorators = []
        for dec in self.decorator_list:
            # Convert decorator to string, replace placeholder and convert back to AST
            dec_str = unparse(dec)
            dec_str = dec_str.replace("/{func_name}", f"/{func_name}")

            # Parse the string back to an AST
            parsed = parse(dec_str).body[0]
            if isinstance(parsed, Expr):
                processed_decorators.append(parsed.value)

        return processed_decorators

    def _build_function_args(self, components: dict[str, Any], func_name: str) -> list[arg]:
        """Generate function arguments from the components/messages field of the AsyncAPI document.

        This function generates the function arguments from the
        components/messages field of the AsyncAPI document. The function
        arguments are generated from the payload properties of the request message.

        Args:
            components (dict): The components field of the AsyncAPI document.
            func_name (str): The name of the function.

        Returns:
            list[arg]: The generated function arguments.
        """
        function_args: list[arg] = []
        request_params = components.get("messages", {}).get(f"{func_name}_request", {}).get("payload", {})

        if request_params.get("properties"):
            for param_name, param_value in request_params["properties"].items():
                annotation_node = convert_asyncapi_property_to_ast_node(param_value)
                if annotation_node and param_name not in request_params.get("required", []):
                    annotation_node = BinOp(
                        left=cast(expr, annotation_node),
                        op=BitOr(),
                        right=Constant(value=None),
                    )
                if annotation_node:
                    function_args.append(arg(arg=param_name, annotation=annotation_node))

        return function_args

    def _build_return_annotation(self, components: dict[str, Any], func_name: str) -> Any:
        """Generate the return annotation of the function from the components/messages field of the AsyncAPI document.

        This function generates the return annotation of the function from the
        components/messages field of the AsyncAPI document. The return annotation
        is generated from the payload properties of the response message.

        Args:
            components (dict): The components field of the AsyncAPI document.
            func_name (str): The name of the function.

        Returns:
            Any: The generated return annotation.
        """
        response = components.get("messages", {}).get(f"{func_name}_response", {})
        response_param = response.get("payload", {})
        returns_node = convert_asyncapi_property_to_ast_node(response_param)

        if response_param and not response_param.get("format") == "required" and returns_node:
            returns_node = BinOp(
                left=cast(expr, returns_node),
                op=BitOr(),
                right=Constant(value=None),
            )

        return returns_node



from __future__ import annotations

from pathlib import Path


class ZenException(Exception):
    pass


class InvalidFile(ZenException):
    def __init__(self, message: str, file_path: Path) -> None:
        """Constructor of InvalidFile exception.

        Args:
            message: Human readable string describing the exception.
            file_path: The path of the invalid file.
        """
        self.message = message
        self.file_path = file_path

    pass



"""This module contains utilities for working with files."""

from __future__ import annotations

import json
from ast import Module, fix_missing_locations, parse, unparse
from pathlib import Path
from typing import Any, Dict, Literal

import yaml

from zen_generator.core.exception import InvalidFile
from zen_generator.core.formatting import format_python_code


def parse_python_file_to_ast(source: Path) -> Module | None:
    """Parse a Python file to its AST.

    Args:
        source: The path of the file to parse.

    Returns:
        The parsed AST or None if the file is a directory or doesn't exist.
    """
    if source.is_file():
        source_text = source.read_text()
        return parse(source_text)
    elif source.is_dir():
        raise InvalidFile("The source is a directory", source)
    elif not source.exists():
        raise InvalidFile("The source doesn't exist", source)
    return None


def load_yaml(source: Path) -> dict[str, Any] | None:
    """Load a YAML file as a dictionary.

    Args:
        source: The path of the file to load.

    Returns:
        The loaded YAML file as a dictionary or None if the file is a directory or
        doesn't exist.
    """
    if source.is_file():
        with open(source, "r") as file:
            return yaml.safe_load(file)
    elif source.is_dir():
        raise InvalidFile("The source is a directory", source)
    elif not source.exists():
        raise InvalidFile("The source doesn't exist", source)
    return None


def write_asyncapi_schema(
    schema: Dict[str, Any],
    output_path: Path,
    format_type: Literal["yaml", "json"] = "yaml",
) -> None:
    """Write an AsyncAPI schema to a file.

    Write the provided AsyncAPI schema to the specified output path in the
    specified format.

    Args:
        schema: The AsyncAPI schema to write.
        output_path: The path of the file to write.
        format_type: The format of the output file. Should be either "yaml" or
            "json". Defaults to "yaml".
    """
    output_path.parent.mkdir(parents=True, exist_ok=True)

    if format_type == "yaml":
        content = yaml.dump(schema, default_flow_style=False)
        file_path = output_path.with_suffix(".yaml")
    else:
        content = json.dumps(schema, indent=2)
        file_path = output_path.with_suffix(".json")

    file_path.write_text(content)
    print(f"Schema AsyncAPI generato: {file_path}")


def save_yaml_file(async_api_content: dict[str, Any] | None, destination: Path, app_name: str) -> None:
    """Save the AsyncAPI content to a file.

    Save the provided AsyncAPI content to the specified destination file or
    directory. If the destination is a directory, the file name will be the
    application name with a .yml extension.

    Args:
        async_api_content: The AsyncAPI content to write.
        destination: The path of the file to write.
        app_name: The name of the application.
    """
    async_api_content = async_api_content or {}
    if destination.is_dir():
        destination = destination / Path(f"{app_name}.yml")

    dumped = yaml.dump(async_api_content, default_flow_style=False, sort_keys=False)
    with open(destination, mode="w") as f:
        f.write(dumped)


def save_python_file(function_body: list[Any], destination: Path) -> None:
    """Save the Python code to a file.

    Save the provided Python code to the specified destination file. The code
    will be formatted according to the PEP 8 style guide.

    Args:
        function_body: The Python code to write.
        destination: The path of the file to write.
    """
    python_module = Module(body=function_body, type_ignores=[])
    python_file_content = unparse(fix_missing_locations(python_module))
    python_file_content = format_python_code(python_file_content)
    with open(destination, mode="w") as f:
        f.write(python_file_content)



from ast import Assign, ClassDef, Constant, Expr, FunctionDef, ImportFrom
from typing import Any
import pytest
from pathlib import Path
from zen_generator.core.ast_utils import generate_code_from_ast
from zen_generator.generators.common_python import BasePythonGenerator


@pytest.fixture
def asyncapi_file(tmp_path) -> Any:
    file = tmp_path / "asyncapi.yaml"
    file.write_text("""
    asyncapi: '2.0.0'
    info:
      title: Test API
      version: '1.0.0'
      description: Test API description
    components:
      schemas:
        TestModel:
          type: object
          properties:
            name:
              type: string
            age:
              type: integer
    """)
    return file


@pytest.fixture
def generator() -> BasePythonGenerator:
    return BasePythonGenerator()


def test_generate_files_from_asyncapi(generator, asyncapi_file, tmp_path):
    models_file = tmp_path / "models.py"
    functions_file = tmp_path / "functions.py"
    generator.generate_files_from_asyncapi(asyncapi_file, models_file, functions_file, "TestApp")

    assert models_file.exists()
    assert functions_file.exists()


def test_add_logger_setup(generator):
    generator._add_logger_setup("TestApp")
    assert any(isinstance(node, Assign) and node.targets[0].id == "logger" for node in generator.functions_ast)


def test_add_docstring(generator):
    generator._add_docstring()
    assert any(
        isinstance(node, Expr) and isinstance(node.value, Constant) and node.value.value == "Test API description"
        for node in generator.functions_ast
    )


def test_add_models_import(generator):
    generator.component_schemas = {"TestModel": {}}
    generator._add_models_import()
    assert any(isinstance(node, ImportFrom) and node.module == ".models" for node in generator.functions_ast)


def test_generate_models_ast(generator):
    generator.component_schemas = {
        "TestModel": {"properties": {"name": {"type": "string"}, "age": {"type": "integer"}}}
    }
    generator.generate_models_ast()
    assert any(isinstance(node, ClassDef) and node.name == "TestModel" for node in generator.models_ast)


def test_generate_function_ast(generator):
    generator.source_content = {
        "components": {"operations": {"test_function": {"description": "Test function description"}}}
    }
    generator.generate_function_ast("TestApp")
    assert any(isinstance(node, FunctionDef) and node.name == "test_function" for node in generator.functions_ast)

    def test_generate_files_from_asyncapi_with_async(generator, asyncapi_file, tmp_path):
        models_file = tmp_path / "models.py"
        functions_file = tmp_path / "functions.py"
        generator.generate_files_from_asyncapi(asyncapi_file, models_file, functions_file, "TestApp", is_async=True)

        assert models_file.exists()
        assert functions_file.exists()
        assert generator.is_async is True


def test_generate_files_from_asyncapi_without_async(generator, asyncapi_file, tmp_path):
    models_file = tmp_path / "models.py"
    functions_file = tmp_path / "functions.py"
    generator.generate_files_from_asyncapi(asyncapi_file, models_file, functions_file, "TestApp", is_async=False)

    assert models_file.exists()
    assert functions_file.exists()
    assert generator.is_async is False


def test_generate_models_ast_with_no_component_schemas(generator):
    generator.component_schemas = {}
    generator.generate_models_ast()
    assert generator.models_ast == []


def test_generate_function_ast_with_no_operations(generator):
    generator.source_content = {"components": {"operations": {}}}
    generator.generate_function_ast("TestApp", logger=False)
    # Even though there are no operations, the functions_ast might not be empty
    # because it can contain a docstring and an import from __future__.
    assert len(generator.functions_ast) == 2
    # Check if the first node is a docstring
    assert isinstance(generator.functions_ast[0], Expr)
    assert isinstance(generator.functions_ast[0].value, Constant)
    assert generator.functions_ast[0].value.value == "Test API description"
    # Check if the second node is an import from __future__
    assert isinstance(generator.functions_ast[1], ImportFrom)
    assert generator.functions_ast[1].module == "__future__"


def test_generate_function_ast_with_logger(generator):
    generator.source_content = {
        "components": {"operations": {"test_function": {"description": "Test function description"}}}
    }
    generator.generate_function_ast("TestApp", logger=True)
    assert any(isinstance(node, Assign) and node.targets[0].id == "logger" for node in generator.functions_ast)


def test_generate_function_ast_without_logger(generator):
    generator.source_content = {
        "components": {"operations": {"test_function": {"description": "Test function description"}}}
    }
    generator.generate_function_ast("TestApp", logger=False)

    assert not any(isinstance(node, Assign) and node.targets[0].id == "logger" for node in generator.functions_ast)
